"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const childProcess = require("child_process");
const crypto = require("crypto");
const fs = require("fs");
const https = require("https");
const path = require("path");
const readline = require("readline");
const zlib = require("zlib");
const mixed_1 = require("../helpers/mixed");
const parse_1 = require("../helpers/parse");
const EMPTY_STDERR = mixed_1.dim("(empty stderr)");
async function download() {
    var _a;
    const parseResult = parse_1.findReadAndParseElmToolingJson();
    switch (parseResult.tag) {
        case "ElmToolingJsonNotFound":
            console.error(parseResult.message);
            return { tag: "Exit", statusCode: 1 };
        case "ReadAsJsonObjectError":
            console.error(mixed_1.bold(parseResult.elmToolingJsonPath));
            console.error(parseResult.message);
            return { tag: "Exit", statusCode: 1 };
        case "Parsed": {
            switch ((_a = parseResult.tools) === null || _a === void 0 ? void 0 : _a.tag) {
                case undefined:
                    console.log(mixed_1.bold(parseResult.elmToolingJsonPath));
                    console.log(`The "tools" field is missing. Nothing to download.`);
                    return { tag: "Exit", statusCode: 0 };
                case "Error":
                    console.error(mixed_1.bold(parseResult.elmToolingJsonPath));
                    console.error("");
                    console.error(parse_1.printFieldErrors(parseResult.tools.errors));
                    console.error("");
                    console.error(mixed_1.elmToolingJsonDocumentationLink);
                    return { tag: "Exit", statusCode: 1 };
                case "Parsed":
                    console.log(mixed_1.bold(parseResult.elmToolingJsonPath));
                    return await downloadTools(parseResult.tools.parsed);
            }
        }
    }
}
exports.default = download;
async function downloadTools(tools) {
    if (tools.existing.length === 0 && tools.missing.length === 0) {
        console.log(`The "tools" field is empty. Nothing to download.`);
        return { tag: "Exit", statusCode: 0 };
    }
    for (const tool of tools.existing) {
        console.log(`${mixed_1.bold(`${tool.name} ${tool.version}`)} already exists: ${mixed_1.dim(tool.absolutePath)}`);
    }
    if (tools.missing.length === 0) {
        return { tag: "Success", tools: tools.existing };
    }
    for (const tool of tools.missing) {
        fs.mkdirSync(path.dirname(tool.absolutePath), { recursive: true });
    }
    const toolsProgress = tools.missing.map(() => 0);
    const firstDrawTime = Date.now();
    const draw = () => {
        console.log(tools.missing
            .map((tool, index) => {
            const progress = toolsProgress[index];
            const progressString = typeof progress === "string"
                ? progress.padEnd(4)
                : Math.round(progress * 100)
                    .toString()
                    .padStart(3) + "%";
            return `${mixed_1.bold(progressString)} ${tool.name} ${tool.version}`;
        })
            .join("\n"));
    };
    const redraw = ({ force = false } = {}) => {
        // Without this time thing, you’ll see the progress go up to 100% and then
        // back to 0% again when using curl. It reports the progress per request –
        // even redirects. Hopefully a redirect response usually finishes in 1 second.
        if (Date.now() - firstDrawTime > 1000 || force) {
            readline.moveCursor(process.stderr, 0, -tools.missing.length);
            draw();
        }
    };
    draw();
    const results = await Promise.all(tools.missing.map((tool, index) => downloadAndExtract(tool, (percentage) => {
        toolsProgress[index] = percentage;
        redraw();
    }).catch((error) => {
        toolsProgress[index] = "ERR!";
        redraw();
        return new Error(downloadAndExtractError(tool, error));
    })));
    const downloadErrors = results.flatMap((result) => result instanceof Error ? result : []);
    redraw({ force: true });
    if (downloadErrors.length > 0) {
        console.error("");
        console.error([
            mixed_1.printNumErrors(downloadErrors.length),
            ...downloadErrors.map((error) => error.message),
        ].join("\n\n"));
        return { tag: "Exit", statusCode: 1 };
    }
    return {
        tag: "Success",
        tools: [...tools.existing, ...tools.missing],
    };
}
function downloadAndExtract(tool, onProgress) {
    return new Promise((resolve, reject) => {
        const removeExtractedAndReject = (error) => {
            hash.destroy();
            extract.destroy();
            download.kill();
            try {
                fs.unlinkSync(tool.absolutePath);
                reject(error);
            }
            catch (removeErrorAny) {
                const removeError = removeErrorAny;
                if (removeError.code === "ENOENT") {
                    reject(error);
                }
                else {
                    reject(new Error(`${error.message}\n${removeError.message}`));
                }
            }
        };
        const hash = crypto.createHash("sha256");
        const extract = extractFile({
            assetType: tool.asset.type,
            file: tool.absolutePath,
            onError: removeExtractedAndReject,
            onSuccess: resolve,
        });
        const download = downloadFile(tool.asset.url, {
            onData: (chunk) => {
                hash.update(chunk);
                extract.write(chunk);
            },
            onProgress,
            onError: removeExtractedAndReject,
            onSuccess: () => {
                const digest = hash.digest("hex");
                if (digest === tool.asset.hash) {
                    extract.end();
                }
                else {
                    removeExtractedAndReject(new Error(hashMismatch(digest, tool.asset.hash)));
                }
            },
        });
    });
}
function downloadAndExtractError(tool, error) {
    return `
${mixed_1.bold(`${tool.name} ${tool.version}`)}
${mixed_1.indent(`
${mixed_1.dim(`< ${tool.asset.url}`)}
${mixed_1.dim(`> ${tool.absolutePath}`)}
${error.message}
  `.trim())}
  `.trim();
}
function hashMismatch(actual, expected) {
    return `
The downloaded file does not have the expected hash!
Expected: ${expected}
Actual:   ${actual}
  `.trim();
}
function downloadFile(url, { onData, onProgress, onError, onSuccess, }) {
    let stderr = "";
    const errored = [];
    const onStderr = (chunk) => {
        stderr += chunk.toString();
        // Extract progress percentage from curl/wget.
        const matches = stderr.match(/\d+(?:[.,]\d+)?%/g) || [];
        if (matches.length > 0) {
            onProgress(Math.min(1, parseFloat(matches[matches.length - 1]) / 100));
        }
    };
    const onClose = (commandName) => (code) => {
        if (errored.includes(commandName)) {
            return;
        }
        else if (code === 0) {
            onSuccess();
        }
        else {
            onError(new Error(`${commandName} exited with non-zero exit code ${code}:\n${stderr
                .trim()
                // Remove curl’s progress bar remnants.
                .replace(/^[\s#O=-]+/g, "") || EMPTY_STDERR}`));
        }
    };
    const curl = childProcess.spawn("curl", ["-#fL", url]);
    let toKill = curl;
    curl.stdout.on("data", onData);
    curl.stderr.on("data", onStderr);
    curl.on("close", onClose("curl"));
    curl.on("error", (error) => {
        errored.push("curl");
        if (error.code === "ENOENT") {
            const wget = childProcess.spawn("wget", ["-O", "-", url]);
            toKill = wget;
            wget.stdout.on("data", onData);
            wget.stderr.on("data", onStderr);
            wget.on("close", onClose("wget"));
            wget.on("error", (error) => {
                errored.push("wget");
                if (error.code === "ENOENT") {
                    toKill = downloadFileNative(url, {
                        onData,
                        onProgress,
                        onError,
                        onSuccess,
                    });
                }
                else {
                    onError(error);
                }
            });
        }
        else {
            onError(error);
        }
    });
    return toKill;
}
function downloadFileNative(url, { onData, onProgress, onError, onSuccess, }, maxRedirects = 50 // This is curl’s default.
) {
    let kill = {
        kill: () => {
            request.destroy();
        },
    };
    const request = https.get(url, (response) => {
        var _a;
        switch (response.statusCode) {
            case 302: {
                const redirect = response.headers.location;
                if (redirect === undefined) {
                    onError(new Error(`Got 302 without location header.`));
                }
                else if (maxRedirects <= 0) {
                    onError(new Error(`Too many redirects.`));
                }
                else {
                    kill = downloadFileNative(redirect, {
                        onData,
                        onProgress,
                        onError,
                        onSuccess,
                    }, maxRedirects - 1);
                }
                break;
            }
            case 200: {
                const contentLength = parseInt(response.headers["content-length"] || "", 10);
                let length = 0;
                response.on("data", (chunk) => {
                    length += chunk.length;
                    onData(chunk);
                    if (Number.isFinite(contentLength) && contentLength > 0) {
                        onProgress(Math.min(1, length / contentLength));
                    }
                });
                response.on("end", onSuccess);
                break;
            }
            default:
                onError(new Error(`Unexpected status code: ${(_a = response.statusCode) !== null && _a !== void 0 ? _a : "unknown"}`));
                break;
        }
    });
    request.on("error", onError);
    return kill;
}
function extractFile({ assetType, file, onError, onSuccess, }) {
    switch (assetType) {
        case "gz": {
            const gunzip = zlib.createGunzip();
            const write = fs.createWriteStream(file);
            gunzip.on("error", onError);
            write.on("error", onError);
            write.on("close", () => {
                // Make executable: `chmod +x`.
                fs.chmod(file, "755", (error) => {
                    if (error === null) {
                        onSuccess();
                    }
                    else {
                        onError(error);
                    }
                });
            });
            gunzip.pipe(write);
            return gunzip;
        }
        case "tgz": {
            const tar = childProcess.spawn("tar", [
                "zxf",
                "-",
                "-C",
                path.dirname(file),
                path.basename(file),
            ]);
            let tarStderr = "";
            tar.on("error", (error) => {
                if (error.code === "ENOENT") {
                    onError(new Error(`tar must be installed on your system and be in $PATH:\n${error.message}`));
                }
                onError(error);
            });
            tar.stderr.on("data", (chunk) => {
                tarStderr += chunk.toString();
            });
            tar.on("close", (code) => {
                if (code === 0) {
                    onSuccess();
                }
                else {
                    onError(new Error(`tar exited with non-zero exit code ${code}:\n${tarStderr.trim() || EMPTY_STDERR}`));
                }
            });
            return tar.stdin;
        }
    }
}
